//-----------------------------------------------------------------------------
// Copyright (c) 2006 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------
//
// This demo illustrates how the GLdriver library is used to load any GL and
// GLU DLLs at runtime. The GLdriver library uses run-time dynamic linking to
// the OpenGL DLLs. This means that the client application no longer needs to
// link to the import libraries for the OpenGL DLLs (i.e., opengl32.lib and
// glu32.lib).
//
// Please note that the GLdriver library only loads the core OpenGL 1.1 to 2.0
// APIs. It does not load any other OpenGL extensions.
//
// The demo itself draws a rotating cube with a procedurally generated checker
// pattern texture.
//
//-----------------------------------------------------------------------------

#include <windows.h>
#include <tchar.h>
#include <mmsystem.h>
#include <string>
#include <vector>
#include "gl_driver.h"

#if defined(_DEBUG)
#include <crtdbg.h>
#endif

#define TEXTURE_WIDTH   64
#define TEXTURE_HEIGHT  64

struct Vertex
{
    float pos[3];
    float normal[3];
    float texcoord[2];
};

// Vertex data for a unit cube centered about the origin.
// Each face contains 2 triangles.
Vertex g_cube[36] =
{
    // Positive Z.
    { 0.5f,  0.5f,  0.5f,   0.0f, 0.0f, 1.0f,   0.0f, 0.0f},
    { 0.5f, -0.5f,  0.5f,   0.0f, 0.0f, 1.0f,   0.0f, 1.0f},
    {-0.5f,  0.5f,  0.5f,   0.0f, 0.0f, 1.0f,   1.0f, 0.0f},
    {-0.5f,  0.5f,  0.5f,   0.0f, 0.0f, 1.0f,   1.0f, 0.0f},
    { 0.5f, -0.5f,  0.5f,   0.0f, 0.0f, 1.0f,   0.0f, 1.0f},
    {-0.5f, -0.5f,  0.5f,   0.0f, 0.0f, 1.0f,   1.0f, 1.0f},

    // Negative Z.
    {-0.5f,  0.5f, -0.5f,   0.0f, 0.0f, -1.0f,   0.0f, 0.0f},
    {-0.5f, -0.5f, -0.5f,   0.0f, 0.0f, -1.0f,   0.0f, 1.0f},
    { 0.5f,  0.5f, -0.5f,   0.0f, 0.0f, -1.0f,   1.0f, 0.0f},
    { 0.5f,  0.5f, -0.5f,   0.0f, 0.0f, -1.0f,   1.0f, 0.0f},
    {-0.5f, -0.5f, -0.5f,   0.0f, 0.0f, -1.0f,   0.0f, 1.0f},
    { 0.5f, -0.5f, -0.5f,   0.0f, 0.0f, -1.0f,   1.0f, 1.0f},

    // Positive Y.
    {-0.5f,  0.5f,  0.5f,   0.0f, 1.0f, 0.0f,   0.0f, 0.0f},
    {-0.5f,  0.5f, -0.5f,   0.0f, 1.0f, 0.0f,   0.0f, 1.0f},
    { 0.5f,  0.5f,  0.5f,   0.0f, 1.0f, 0.0f,   1.0f, 0.0f},
    { 0.5f,  0.5f,  0.5f,   0.0f, 1.0f, 0.0f,   1.0f, 0.0f},
    {-0.5f,  0.5f, -0.5f,   0.0f, 1.0f, 0.0f,   0.0f, 1.0f},
    { 0.5f,  0.5f, -0.5f,   0.0f, 1.0f, 0.0f,   1.0f, 1.0f},

    // Negative Y.
    {-0.5f, -0.5f, -0.5f,   0.0f, -1.0f, 0.0f,   0.0f, 0.0f},
    {-0.5f, -0.5f,  0.5f,   0.0f, -1.0f, 0.0f,   0.0f, 1.0f},
    { 0.5f, -0.5f, -0.5f,   0.0f, -1.0f, 0.0f,   1.0f, 0.0f},
    { 0.5f, -0.5f, -0.5f,   0.0f, -1.0f, 0.0f,   1.0f, 0.0f},
    {-0.5f, -0.5f,  0.5f,   0.0f, -1.0f, 0.0f,   0.0f, 1.0f},
    { 0.5f, -0.5f,  0.5f,   0.0f, -1.0f, 0.0f,   1.0f, 1.0f},

    // Positive X.
    { 0.5f,  0.5f, -0.5f,   1.0f, 0.0f, 0.0f,   0.0f, 0.0f},
    { 0.5f, -0.5f, -0.5f,   1.0f, 0.0f, 0.0f,   0.0f, 1.0f},
    { 0.5f,  0.5f,  0.5f,   1.0f, 0.0f, 0.0f,   1.0f, 0.0f},
    { 0.5f,  0.5f,  0.5f,   1.0f, 0.0f, 0.0f,   1.0f, 0.0f},
    { 0.5f, -0.5f, -0.5f,   1.0f, 0.0f, 0.0f,   0.0f, 1.0f},
    { 0.5f, -0.5f,  0.5f,   1.0f, 0.0f, 0.0f,   1.0f, 1.0f},

    // Negative X.
    {-0.5f,  0.5f,  0.5f,   -1.0f, 0.0f, 0.0f,   0.0f, 0.0f},
    {-0.5f, -0.5f,  0.5f,   -1.0f, 0.0f, 0.0f,   0.0f, 1.0f},
    {-0.5f,  0.5f, -0.5f,   -1.0f, 0.0f, 0.0f,   1.0f, 0.0f},
    {-0.5f,  0.5f, -0.5f,   -1.0f, 0.0f, 0.0f,   1.0f, 0.0f},
    {-0.5f, -0.5f,  0.5f,   -1.0f, 0.0f, 0.0f,   0.0f, 1.0f},
    {-0.5f, -0.5f, -0.5f,   -1.0f, 0.0f, 0.0f,   1.0f, 1.0f}
};

HWND  g_hWnd;
HDC   g_hDC;
HGLRC g_hRC;
INT   g_width;
INT   g_height;
INT   g_frames;
UINT  g_textureId;
GLdriver g_driver;

void Cleanup()
{
    if (g_textureId)
    {
        glDeleteTextures(1, &g_textureId);
        g_textureId = 0;
    }

    if (g_hDC)
    {
        if (g_hRC)
        {
            wglMakeCurrent(g_hDC, 0);
            wglDeleteContext(g_hRC);
            g_hRC = 0;
        }

        ReleaseDC(g_hWnd, g_hDC);
        g_hDC = 0;
    }

    // Unloads the OpenGL DLLs and NULLs out all the OpenGL API function
    // pointers. Call this *after* the rendering context has been destroyed.
    GLdriverUnload(&g_driver);
}

void InitCheckerPatternTexture()
{
    // Procedurally create a black and white checker pattern texture.
    // Adapted from the "OpenGL Programming Guide" (aka the red book).

    BYTE *pPixels = 0;
    BYTE c = 0;
    UINT pitch = TEXTURE_WIDTH * 4;
    std::vector<BYTE> checkerImage(pitch * TEXTURE_HEIGHT);

    for (UINT i = 0; i < TEXTURE_HEIGHT; ++i)
    {
        for (UINT j = 0; j < TEXTURE_WIDTH; ++j)
        {
            c = (BYTE)((((i & 0x8) == 0) ^ ((j & 0x8) == 0)) * 255);
            pPixels = &checkerImage[i * pitch + (j * 4)];
            pPixels[0] = pPixels[1] = pPixels[2] = c;
            pPixels[3] = 255;
        }
    }

    glGenTextures(1, &g_textureId);
    glBindTexture(GL_TEXTURE_2D, g_textureId);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    glTexEnvf(GL_TEXTURE_2D, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH, TEXTURE_HEIGHT,
        0, GL_BGRA, GL_UNSIGNED_BYTE, &checkerImage[0]);
}

bool Init(LPCTSTR lpszPathGL, LPCTSTR lpszPathGLU)
{
    // Manually load the GL and GLU DLLs.
    // Once the DLLs are loaded *only* OpenGL 1.1 functions will be available.
    // OpenGL beyond version 1.1 is only available *after* a rendering context
    // has been activated and *after* GLdriverInit() is called.
    if (!GLdriverLoad(&g_driver, lpszPathGL, lpszPathGLU))
        return false;

    PIXELFORMATDESCRIPTOR pfd = {0};

    pfd.nSize = sizeof(pfd);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 24;
    pfd.cDepthBits = 16;
    pfd.iLayerType = PFD_MAIN_PLANE;

    if (!(g_hDC = GetDC(g_hWnd)))
        return false;

    int pf = ChoosePixelFormat(g_hDC, &pfd);

    if (!SetPixelFormat(g_hDC, pf, &pfd))
        return false;

    if (!(g_hRC = wglCreateContext(g_hDC)))
        return false;

    if (!wglMakeCurrent(g_hDC, g_hRC))
        return false;

    // Once the OpenGL rendering context has been activated all of the OpenGL
    // API functions (greater than OpenGL 1.1) must be explicitly retrieved
    // from the loaded OpenGL DLLs.
    GLdriverInit(&g_driver);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_TEXTURE_2D);

    InitCheckerPatternTexture();
    glBindTexture(GL_TEXTURE_2D, g_textureId);

    return true;
}

void RenderFrame()
{
    glViewport(0, 0, g_width, g_height);
    glClearColor(0.0f, 0.0f, 0.8f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (float)g_width / (float)g_height, 1.0f, 100.0f);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(0.0f, 0.0f, 3.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);

    // This seems to produce a nice smooth rotation for our cube.
    glRotatef(timeGetTime() / 50.0f, 0.0f, 1.0f, 0.0f);

    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_VERTEX_ARRAY);   

    glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), g_cube[0].texcoord);
    glNormalPointer(GL_FLOAT, sizeof(Vertex), g_cube[0].normal);
    glVertexPointer(3, GL_FLOAT, sizeof(Vertex), g_cube[0].pos);

    glDrawArrays(GL_TRIANGLES, 0, sizeof(g_cube) / sizeof(g_cube[0]));

    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);

    SwapBuffers(g_hDC);
    g_frames++;
}

LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static TCHAR szBuffer[32] = {0};

    switch (msg)
    {
    case WM_CREATE:
        timeBeginPeriod(1);
        SetTimer(hWnd, 1, 1000, 0);
        return 0;

    case WM_DESTROY:
        KillTimer(hWnd, 1);
        timeEndPeriod(1);
        PostQuitMessage(0);
        return 0;

    case WM_KEYDOWN:
        if (wParam == VK_ESCAPE)
        {
            SendMessage(hWnd, WM_CLOSE, 0, 0);
            return 0;
        }
        break;

    case WM_SIZE:
        g_width = (INT)LOWORD(lParam);
        g_height = (INT)HIWORD(lParam);
        return 0;

    case WM_TIMER:
        _stprintf(szBuffer, _T("%d FPS"), g_frames);
        SetWindowText(hWnd, szBuffer);
        g_frames = 0;
        return 0;

    default:
        break;
    }

    return DefWindowProc(hWnd, msg, wParam, lParam);
}

void ResizeAndCenterWindow()
{
    // Resize and center the created window so that it is exactly a quarter the
    // size of the current desktop's dimensions.

    int w = GetSystemMetrics(SM_CXSCREEN) / 2;
    int h = GetSystemMetrics(SM_CYSCREEN) / 2;
    UINT screenWidth = GetSystemMetrics(SM_CXSCREEN);
    UINT screenHeight = GetSystemMetrics(SM_CYSCREEN);
    UINT left = (screenWidth - w) / 2;
    UINT top = (screenHeight - h) / 2;
    RECT rcWindow;

    SetRect(&rcWindow, left, top, left + w, top + h);
    
    AdjustWindowRectEx(&rcWindow,
        WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS, FALSE, 0);

    MoveWindow(g_hWnd, rcWindow.left, rcWindow.top,
        rcWindow.right - rcWindow.left, rcWindow.bottom - rcWindow.top, TRUE);
}

#if defined(UNICODE) && defined(_UNICODE)
std::wstring ToUnicodeString(const std::string str)
{
    std::wstring wstr;
    const std::ctype<wchar_t> &facet = std::use_facet<std::ctype<wchar_t> >(std::locale());
    std::string::const_iterator i = str.begin();
    std::string::const_iterator end = str.end();

    while (i != end)
    {
        wstr += facet.widen(*i);
        ++i;
    }

    return wstr;
}
#endif

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpCmdLine, int nShowCmd)
{
#if defined(_DEBUG)
    _CrtSetDbgFlag(_CRTDBG_LEAK_CHECK_DF | _CRTDBG_ALLOC_MEM_DF);
    _CrtSetReportMode(_CRT_ASSERT, _CRTDBG_MODE_FILE);
    _CrtSetReportFile(_CRT_ASSERT, _CRTDBG_FILE_STDERR);
#endif

    MSG msg = {0};
    WNDCLASSEX wcl = {0};

    wcl.cbSize = sizeof(wcl);
    wcl.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
    wcl.lpfnWndProc = WindowProc;
    wcl.cbClsExtra = 0;
    wcl.cbWndExtra = 0;
    wcl.hInstance = hInstance;
    wcl.hIcon = 0;
    wcl.hCursor = LoadCursor(0, IDC_ARROW);
    wcl.hbrBackground = 0;
    wcl.lpszMenuName = 0;
    wcl.lpszClassName = _T("GLWindowClass");
    wcl.hIconSm = 0;

    if (!RegisterClassEx(&wcl))
        return 0;

    g_hWnd = CreateWindowEx(0, wcl.lpszClassName, _T("GLdriver"),
                WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
                0, 0, 0, 0, 0, 0, wcl.hInstance, 0);

    if (g_hWnd)
    {
        LPCTSTR lpszPathGL = 0;
        LPCTSTR lpszPathGLU = 0;

    #if defined(_UNICODE)
        std::wstring unicodePathGL;
        std::wstring unicodePathGLU;
    #endif

        // Process the command-line arguments.
        switch (__argc)
        {
        case 1:
            lpszPathGL = _T("opengl32.dll");
            lpszPathGLU = _T("glu32.dll");
            break;

        case 2:
            #if defined(UNICODE) && defined(_UNICODE)
                unicodePathGL = ToUnicodeString(__argv[1]);
                lpszPathGL = unicodePathGL.c_str();
            #else
                lpszPathGL = __argv[1];
            #endif
            break;

        case 3:
            #if defined(UNICODE) && defined(_UNICODE)
                unicodePathGL = ToUnicodeString(__argv[1]);
                unicodePathGLU = ToUnicodeString(__argv[2]);
                lpszPathGL = unicodePathGL.c_str();
                lpszPathGLU = unicodePathGLU.c_str();
            #else
                lpszPathGL = __argv[1];
                lpszPathGLU = __argv[2];
            #endif
            break;

        default:
            MessageBox(0, 
                _T("Usage:\n")
                _T("GLdriver [<GL DLL path> <GLU DLL path>]\n\n")
                _T("If no args are supplied the default system GL is used.\n"),
                _T("GLdriver Usage"), MB_OK);

            return 0;
        }

        if (Init(lpszPathGL, lpszPathGLU))
        {
            ResizeAndCenterWindow();

            ShowWindow(g_hWnd, nShowCmd);
            UpdateWindow(g_hWnd);

            while (true)
            {
                if (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
                {
                    if (msg.message == WM_QUIT)
                        break;

                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }
                else
                {
                    RenderFrame();
                }
            }
        }

        Cleanup();
        UnregisterClass(wcl.lpszClassName, hInstance);
    }

    return (int)(msg.wParam);
}